/* 
   bmcfenced.c
   ( Originally called portblocker.c )

   Copyright (C) 2003-2004 FreeIPMI Core Team

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

/* Modification History:
 * 21-May-2004: Anand Manian : Changed original name (portblocker.c)
 * 21-May-2004: Anand Manian : Modified to handle multiple ports
*/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/ioctl.h>

#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>

char pidfile[128];
struct sockaddr_in myaddr;   // server address

void 
tidy_up (int mysig)
{
  //fprintf (stderr, "shutting down\n");
  unlink (pidfile);
  exit (0);
}

int
accept_connection (int listener)
{
  struct sockaddr_in remoteaddr; // client address
  socklen_t addrlen;
  int newfd;
  
  // handle new connections
  addrlen = sizeof (remoteaddr);
  newfd = accept (listener, (struct sockaddr *) &remoteaddr, 
		  &addrlen);
  
  if (newfd == -1)
    {
      perror ("bmcfenced");
      return -1;
    }
  
  write (newfd, "DO NOT CONNECT TO ME\n", strlen ("DO NOT CONNECT TO ME\n"));
  close (newfd);
  
  return 0;
}

/* Spawn process to block a port:
 * Return value:
 *      0   : Spawn was successful and am the new process
 *     -1   : Ran into some error condition during spawn
 *     > 0  : Spawn successful; am the parent and value returned
 *            is PID of newly spawned blocker
 *
 */

int 
daemon_init (int server_socket)
{
  pid_t pid, thispid;
  FILE  *pidout;
  int fd;
  
  if ((pid = fork ()) < 0) 
    {
      perror("fork()");
      return -1;
    }
  else 
    {
      if (pid != 0) 
	return (pid);   /* exit the original foreground call */
    }
  
  /* Only the child process continues */
  setsid ();
  umask (0);
  
  //errout = fopen("/var/log/usb_perm.log", "a");
  //  fprintf((char *)errout,"----------\n");
  
  thispid = getpid ();
  pidout = fopen (pidfile, "w");
  if (pidout == NULL)
  {
    perror("Error opening %s in daemon_init()");
    return -1;
  }
  
  fprintf (pidout, "%d\n", thispid);
  fclose (pidout);
  
  if ((fd = open ("/dev/tty", O_RDWR)) != -1)
    {
      ioctl (fd, TIOCNOTTY, (char *) 0);
      close (fd);
    }
  
  signal (SIGTERM, tidy_up);
  signal (SIGINT, tidy_up);
  
  //signal (SIGTSTP, tidy_up);
  //fprintf (stderr, "daemon started.\n");
  
  while (1) 
    accept_connection (server_socket);

  perror("Unexpected exit by PID=%d");

  return 0;  /*... This should never get executed.. right? ..*/
}

int 
init_server (int port)
{
  int yes = 1;                 // for setsockopt() SO_REUSEADDR, below
  int listener = 0;            // listening socket descriptor
  
  // get the listener
  if ((listener = socket (AF_INET, SOCK_STREAM, 0)) == -1)
    {
      perror ("socket()");
      return (-1);
    }
  
  // lose the pesky "address already in use" error message
  if (setsockopt (listener, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (int)) 
      == -1)
    {
      perror("setsockopt()");
      return (-1);
    }
  
  // bind
  myaddr.sin_family = AF_INET;
  myaddr.sin_addr.s_addr = INADDR_ANY;
  myaddr.sin_port = htons (port);
  memset (&(myaddr.sin_zero), '\0', 8);
  if (bind (listener, (struct sockaddr *)&myaddr, sizeof(myaddr)) == -1)
    {
      perror ("bind()");
      return (-1);
    }
  
  // listen
  if (listen (listener, 10) == -1)
    {
      perror ("listen()");
      return (-1);
    }

    return listener;
}

void
Usage()
{
  fprintf (stderr, "Usage: %s port-to-block ...\n", program_invocation_short_name);
  exit (-1);
}

int 
main (int argc, char **argv)
{
  int server_socket;
  int port2block, argv_indx, daemon_init_status, ret_status=0;

  if (argc < 2) Usage();
  
  for (argv_indx=1; argv_indx < argc; argv_indx++)
  {
     if ( (port2block=atoi(argv[argv_indx])) < 1 )
     {  fprintf(stderr, "%s: Invalid port number !%s!\n",
  	     program_invocation_short_name, argv[argv_indx]); 
     }
     else
     {
        server_socket=init_server(port2block);
	if (server_socket <=0 )
	{       
          fprintf (stderr, "%s: init_server() failed for port %d\n",
      	     program_invocation_short_name,port2block);
          ret_status--;
	}
	else
	{
          sprintf (pidfile, "/var/run/bmcfenced%d.pid", port2block);
  	  daemon_init_status=daemon_init(server_socket);

          if (daemon_init_status < 0 ) /*.. something wrong ..*/
          {
            fprintf(stderr,
               "%s: daemon_init() failed for port %d\n",
  	        program_invocation_short_name, port2block);
  	    ret_status--;
          }
  
          if (daemon_init_status > 0) /*.. parent process ..*/
          {
  	    fprintf(stdout,
  		 "%s: Holding port %d open (PID=%d)\n",
  		 program_invocation_short_name, port2block,
  		 daemon_init_status);
  	  }
  	  else   /*.. Spawned child; We should never get here ..*/
  	    fprintf(stderr,
  			"%s: Port %d no longer held open\n",
  	  	    program_invocation_short_name, port2block);
	}
     }
  }
  return ret_status;
}

